@@ -16,69 +16,78 @@ Description:
 History:
 	960110	created
 //////////////////////////////////////////////////////////////////////#*/
+#include <stdlib.h>
+#include "ystring.h"
 #include "delegate.h"
+#include "fpoll.h"
+#include "file.h"
 #include "http.h"
+#include "proc.h"
 #include <ctype.h>
 #include <errno.h>
 #define MY_CGIVER	"1.1"
 
-extern int main_argc;
-extern char **main_argv;
+int scan_SHTML(Connection *Conn,FILE *tc,FILE *fc,FILE *fp,PCStr(req),PCStr(head),PCStr(vurl),PCStr(ourl),PCStr(path),PCStr(script),PCStr(expath));
+int file_copyTimeout(FILE *src,FILE *dst,FILE *cache,int bytes,int *binary,int timeout);
+
 extern char **environ;
-extern char *HTTP_outCharset();
-extern char *getFieldValue2();
 #define getFieldValue(str,fld,buf,siz) getFieldValue2(str,fld,buf,siz)
-#define getFV(str,fld,buf)             getFieldValue2(str,fld,buf,sizeof(buf))
+#define getFV(str,fld,buf)             getFieldValue2(str,fld,AVStr(buf),sizeof(buf))
 
-addList(list,lsize,elem)
-	char *list,*elem;
-{	char *tp;
 
+void addList(PVStr(list),int lsize,PCStr(elem))
+{	refQStr(tp,list); /**/
 	tp = strtail(list);
+
 	if( list < tp ){
-		linescanX(", ",tp,lsize-(tp-list));
+		linescanX(", ",AVStr(tp),lsize-(tp-list));
 		tp += strlen(tp);
 	}
-	linescanX(elem,tp,lsize-(tp-list));
+	linescanX(elem,AVStr(tp),lsize-(tp-list));
 }
 
-static cgi_head2env(head,Evp)
-	unsigned char *head;
-	StrVec *Evp;
-{	char field[0x1000];
-	char accepts[2048],langs[256];
-	unsigned char *fp,*np,*vp;
+static void cgi_head2env(PCStr(head),StrVec *Evp)
+{	CStr(field,0x1000);
+	CStr(accepts,2048);
+	CStr(langs,256);
+	const char *np;
+	const char *fp;
+	const char *vp; /* not "const" but fixed */
 
 	accepts[0] = 0;
 	langs[0] = 0;
 
-	Strncpy(field,head,sizeof(field)-2);
+	QStrncpy(field,head,sizeof(field)-2);
 	strcat(field,"\r\n");
 	fp = field;
 
 	while( np = strpbrk(fp,"\r\n") ){
-		*np++ = 0;
+		truncVStr(np); np++;
 		for( vp = fp; *vp; vp++ ){
+			if( *vp == ' ' ){
+				goto NEXTFIELD;
+			}
 			if( *vp == ':' ){
-				*vp++ = 0;
+				truncVStr(vp); vp++;
 				if( isspace(*vp) )
 					vp++;
 				break;
 			}
 			if( islower(*vp) )
-				*vp = toupper(*vp);
+				*(char*)vp = toupper(*vp);
 			else
 			if( *vp == '-' )
-				*vp = '_';
+				*(char*)vp = '_';
 		}
 		if( strcmp(fp,"ACCEPT_LANGUAGE") == 0 ){
-			addList(langs,sizeof(langs),vp);
+			addList(AVStr(langs),sizeof(langs),vp);
 		}else
 		if( strcmp(fp,"ACCEPT") == 0 ){
-			addList(accepts,sizeof(accepts),vp);
+			addList(AVStr(accepts),sizeof(accepts),vp);
 		}else{
 			SVaddEnvf(Evp,"HTTP_%s=%s",fp,vp);
 		}
+	NEXTFIELD:
 		fp = np;
 		while( *fp == '\r' || *fp == '\n' )
 			fp++;
@@ -92,44 +101,45 @@ static cgi_head2env(head,Evp)
 	}
 }
 
-static char *cgienv;
-scan_CGIENV(Conn,envlist)
-	Connection *Conn;
-	char *envlist;
+static const char *cgienv;
+int scan_CGIENV(Connection *Conn,PCStr(envlist))
 {
 	cgienv = StrAlloc(envlist);
 	return 0;
 }
 
-cgi_makeEnv(conninfo,req,head,vurl,vpath,datapath,scripturl,extrapath,av,Evp)
-	char *conninfo;
-	char *req,*head;
-	char *vurl,*vpath,*datapath,*scripturl,*extrapath;
-	char *av[];
-	StrVec *Evp;
-{	char *search;
-	char *dp;
-	char tmp[2048];
+void cgi_makeEnv(PCStr(conninfo),PCStr(req),PCStr(head),PCStr(vurl),PCStr(vpath),PCStr(datapath),PCStr(scripturl),PCStr(extrapath),int mac,const char *av[],StrVec *Evp)
+{	const char *search;
+	const char *dp;
+	CStr(tmp,2048);
 	int ac,ei;
-	char *es;
-	char *fp,*np,*vp;
-	char auth[1024],atype[128],auserpass[256],auser[256];
-	char method[128];
-	char *randenv;
+	const char *es;
+	const char *fp;
+	const char *np;
+	const char *vp;
+	CStr(auth,1024);
+	CStr(atype,128);
+	CStr(auserpass,256);
+	CStr(auser,256);
+	CStr(method,128);
+	const char *randenv;
 
 	randenv = 0;
 	for( ei = 0; es = environ[ei]; ei++ ){
 		if( strncmp(es,"RANDENV=",8) == 0 )
 			randenv = es;
 		else
-		if( cgienv == NULL || strmatch_list(es,cgienv,"^=",NULL,NULL) )
-			Evp->sv_ev[Evp->sv_ec++] = es;
+		if( cgienv == NULL || strmatch_list(es,cgienv,"^=",NULL,NULL) ){
+			if( Evp->sv_ecmax-1 <= Evp->sv_ec )
+				syslog_ERROR("CGIENV OVERFLOW %d/%d\n",Evp->sv_ec,Evp->sv_ecmax);
+			else	Evp->sv_ev[Evp->sv_ec++] = (char*)es;
+		}
 	}
 
 	atype[0] = auser[0] = 0;
 	if( getFV(head,"Authorization",auth) ){
-		HTTP_decompAuth(auth,atype,sizeof(atype),auserpass,sizeof(auserpass));
-		sscanf(auserpass,"%[^:]",auser);
+		HTTP_decompAuth(auth,AVStr(atype),sizeof(atype),AVStr(auserpass),sizeof(auserpass));
+		Xsscanf(auserpass,"%[^:]",AVStr(auser));
 	}
 
 	/* AUTH_TYPE */
@@ -156,28 +166,30 @@ cgi_makeEnv(conninfo,req,head,vurl,vpath,datapath,scripturl,extrapath,av,Evp)
 	SVaddEnvf(Evp,"PATH_INFO=%s",extrapath);
 	if( es )
 	if( search = strchr(es,'?') )
-		*search = 0;
+		truncVStr(search);
 
 	/* PATH_TRANSLATED */
 	SVaddEnvf(Evp,"PATH_TRANSLATED=%s",*extrapath?datapath:"");
 	Verbose("PATH_TRANSLATED=%s\n",*extrapath?datapath:"");
 
 	/* QUERY STRING */
-	if( search = strchr(vpath,'?') )
-		*search++ = 0;
+	if( search = strchr(vpath,'?') ){
+		truncVStr(search);
+		search++;
+	}
 	SVaddEnvf(Evp,"QUERY_STRING=%s",search?search:"");
 
 	/* REMOTE_ADDR */
 	/* REMOTE_HOST */
 	/* REMOTE_IDENT */
 	/* REMOTE_USER */
-	getFieldValue(conninfo,"Client-Addr",tmp,sizeof(tmp));
+	getFieldValue(conninfo,"Client-Addr",AVStr(tmp),sizeof(tmp));
 	SVaddEnvf(Evp,"REMOTE_ADDR=%s",tmp);
 
-	getFieldValue(conninfo,"Client-Host",tmp,sizeof(tmp));
+	getFieldValue(conninfo,"Client-Host",AVStr(tmp),sizeof(tmp));
 	SVaddEnvf(Evp,"REMOTE_HOST=%s",tmp);
 
-	getFieldValue(conninfo,"Client-User-Ident",tmp,sizeof(tmp));
+	getFieldValue(conninfo,"Client-User-Ident",AVStr(tmp),sizeof(tmp));
 	if( strcmp(tmp,"-") == 0 )
 		tmp[0] = 0;
 	SVaddEnvf(Evp,"REMOTE_IDENT=%s",tmp);
@@ -198,16 +210,22 @@ cgi_makeEnv(conninfo,req,head,vurl,vpath,datapath,scripturl,extrapath,av,Evp)
 	/* SERVER_PORT */
 	/* SERVER_PROTOCOL */
 	/* SERVER_SOFTWARE */
-	{	char svhp[256],svhost[256];
+	{	CStr(svhp,256);
+		CStr(svhost,256);
 		int svport;
 
 		svhost[0] = 0;
 		if( getFV(head,"Host",svhp) ){
 			svport = 80;
-			sscanf(svhp,"%[^:]:%d",svhost,&svport);
+			Xsscanf(svhp,"%[^:]:%d",AVStr(svhost),&svport);
 		}
 		if( svhost[0] == 0 ){
-			GetHostname(svhost,sizeof(svhost));
+			getFieldValue(conninfo,"Client-IF-Host",AVStr(svhp),sizeof(svhp));
+			Xsscanf(svhp,"%[^:]:%d",AVStr(svhost),&svport);
+			svport = scan_hostport("http",svhp,AVStr(svhost));
+		}
+		if( svhost[0] == 0 ){
+			GetHostname(AVStr(svhost),sizeof(svhost));
 			svport = SERVER_PORT();
 		}
 
@@ -218,7 +236,7 @@ cgi_makeEnv(conninfo,req,head,vurl,vpath,datapath,scripturl,extrapath,av,Evp)
 				DELEGATE_ver());
 	}
 	if( randenv )
-		Evp->sv_ev[Evp->sv_ec++] = randenv;
+		Evp->sv_ev[Evp->sv_ec++] = (char*)randenv;
 	Evp->sv_ev[Evp->sv_ec] = 0;
 
 	if( av != NULL ){
@@ -226,8 +244,11 @@ cgi_makeEnv(conninfo,req,head,vurl,vpath,datapath,scripturl,extrapath,av,Evp)
 		if( search ){
 			fp = SPrintf(Evp->sv_MemF,"%s+",search);
 			while( np = strchr(fp,'+') ){
-				av[ac++] = fp;
-				*np = 0;
+				if( mac-1 <= ac ){
+					break;
+				}
+				av[ac++] = (char*)fp;
+				truncVStr(np);
 				fp = np + 1;
 			}
 		}
@@ -239,21 +260,21 @@ for( ei = 0; ei < ec; ei++ ) fprintf(stderr,"#### ENV[%2d] %s\n",ei,ev[ei]);
 */
 }
 
-static cgi_response(Conn,req,ihead,in,out,stcodep)
-	Connection *Conn;
-	char *req,*ihead;
-	FILE *in,*out;
-	int *stcodep;
-{	char ohead[0x10000],*hp;
-	char ctype[1024],status[1024];
+static int cgi_response(Connection *Conn,PCStr(req),PVStr(ihead),FILE *in,FILE *out,FILE **xout,int *stcodep)
+{	CStr(ohead,0x10000);
+	refQStr(hp,ohead); /**/
+	CStr(ctype,1024);
+	CStr(status,1024);
 	int status_set;
-	char location[1024];
-	char mimever[1024];
-	char field[1024],value[1024];
-	char line[1024],*tp;
+	CStr(location,1024);
+	CStr(mimever,1024);
+	CStr(field,1024);
+	CStr(value,1024);
+	CStr(line,1024);
+	const char *tp;
 	int hleng,bleng,cleng,hcc;
-	char ostat[1024];
-	char *xcharset;
+	CStr(ostat,1024);
+	const char *xcharset;
 	int codeconv;
 	int headonly;
 	int putConnection;
@@ -264,10 +285,9 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 		Conn = &ConnBuff;
 	}
 
-	hp = ohead;
 	xcharset = HTTP_outCharset(Conn);
 
-	if( getFieldValue2(ihead,"X-Status",status,sizeof(status)) )
+	if( getFieldValue2(ihead,"X-Status",AVStr(status),sizeof(status)) )
 	{	HttpResponse Stat;
 	 	if( 0 < decomp_http_status(status,&Stat) ){
 			sprintf(status,"%d %s (filtered by CFI)",
@@ -280,7 +300,7 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 	location[0] = 0;
 	status_set = 0;
 	if( getFV(ihead,"Content-Type",ctype) )
-		rmField(ihead,"Content-Type");
+		rmField(AVStr(ihead),"Content-Type");
 	else	strcpy(ctype,"text/plain");
 	strcpy(mimever,MY_MIMEVER);
 
@@ -295,7 +315,7 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 		hcc += strlen(line);
 
 		if( tp = strpbrk(line,"\r\n") )
-			*tp = 0;
+			truncVStr(tp);
 
 		if( line[0] == 0 )
 			break;
@@ -345,12 +365,12 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 	}
 
 	if( location[0] ){
-		char server[256];
+		CStr(server,256);
 		if( status_set == 0 )
 		strcpy(status,"302 Moved (output of CGI)");
 		if( location[0] == '/' ){
 			if( Conn )
-				ClientIF_HP(Conn,server);
+				ClientIF_HP(Conn,AVStr(server));
 			else{
 				sprintf(server,"%s:%s",
 					getenv("SERVER_NAME"),
@@ -364,7 +384,7 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 	sprintf(hp,"Content-Type: %s\r\n",ctype);
 	codeconv = strncasecmp(ctype,"text/",5) == 0 && xcharset != NULL;
 	if( codeconv )
-		replace_charset(hp,xcharset);
+		replace_charset(AVStr(hp),xcharset);
 	hp += strlen(hp);
 
 	if( atoi(status) == 304 )
@@ -376,7 +396,7 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 	if( headonly
 	 || 0 < cleng && file_size(fileno(in)) - ftell(in) == cleng ){
 		if( Conn ){
-			if( getKeepAlive(Conn,hp) ){
+			if( getKeepAlive(Conn,AVStr(hp)) ){
 				putConnection = 1;
 				hp += strlen(hp);
 			}
@@ -393,6 +413,16 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 	hp += strlen(hp);
 	strcpy(hp,"\r\n");
 
+	/*
+	if( strcaseeq(ctype,"text/shtml") ){
+	*/
+	if( strncasecmp(ctype,"text/shtml",10) == 0 ){
+		if( xout != 0 ){
+			*xout = TMPFILE("CGI -> SHTML");
+			out = *xout;
+		}
+	}
+
 	fputs(ostat,out);
 	fputs(ohead,out);
 	hleng = strlen(ostat) + strlen(ohead);
@@ -409,42 +439,43 @@ static cgi_response(Conn,req,ihead,in,out,stcodep)
 	}
 	return hleng+bleng;
 }
-static putExecError(tcd,execpath)
-	char *execpath;
+static void putExecError(int tcd,PCStr(execpath))
 {	FILE *tc;
-	char *env;
+	const char *env;
+	int serrno = errno;
 
 	sv1log("#### FAILED EXEC: %s\n",execpath);
 	if( env = getenv("PATH") )
 		sv1log("#### PATH=%s\n",env);
 
 	tc = fdopen(tcd,"w");
+	if( serrno == ENOENT )
+	fprintf(tc,"Status: 404 not found\r\n");
+	else
 	fprintf(tc,"Status: 500 cannot execute\r\n");
 	fprintf(tc,"Content-Type: text/plain\r\n");
 	fprintf(tc,"\r\n");
-	fprintf(tc,"Couldn't find or execte the CGI script.\r\n");
+	fprintf(tc,"Couldn't find or execute the CGI script.\r\n");
 }
-cgi_process(Conn,tc,execpath,workdir,av,ev,pfp)
-	Connection *Conn;
-	FILE *tc;
-	char *execpath,*workdir;
-	char *av[],*ev[];
-	FILE *pfp[];
+extern const char *BINSHELL;
+int cgi_process(Connection *Conn,FILE *tc,PCStr(execpath),PCStr(workdir),const char *av[],const char *ev[],FILE *pfp[])
 {	int toCGI[2],fromCGI[2];
-	char **savenv;
-	char savdir[1024];
-	char savfds[2];
+	const char *const *savenv;
+	CStr(savdir,1024);
+	CStr(savfds,2);
 	int pid;
 	FILE *cfp;
-	char *cav[32],cab[1024],line[64];
+	const char *cav[32]; /**/
+	CStr(cab,1024);
 	int cai;
+	int ai;
 
 	pipe(toCGI);
 	pipe(fromCGI);
-	setCloseOnExec(toCGI[1],1);
-	setCloseOnExec(fromCGI[0],1);
-	savenv = environ;
-	environ = ev;
+	setCloseOnExec(toCGI[1] /*,1*/);
+	setCloseOnExec(fromCGI[0] /*,1*/);
+	savenv = (char const*const*)environ;
+	environ = (char**)ev;
 
 	sv1log("chdir(%s)\n",workdir);
 	getcwd(savdir,sizeof(savdir));
@@ -454,7 +485,11 @@ cgi_process(Conn,tc,execpath,workdir,av,ev,pfp)
 	savfds[1] = dup(1); dup2(fromCGI[1],1); close(fromCGI[1]);
 
 	if( cfp = fopen(execpath,"r") ){
-		fgets(line,sizeof(line),cfp);
+		CStr(line,64);
+		CStr(shell,32);
+		bzero(line,sizeof(line));
+		fread(line,1,sizeof(line),cfp);
+		wordscanX(line,AVStr(shell),sizeof(shell)-1);
 		if( strncasecmp(line,"#!CGI-DeleGate",14) == 0 ){
 /*
  * should pass access control parameters to control the access via CGI...
@@ -462,19 +497,34 @@ cgi_process(Conn,tc,execpath,workdir,av,ev,pfp)
 			sprintf(cab,"+=%s",execpath);
 			execpath = EXEC_PATH;
 			cai = 0;
-			cav[cai++] = execpath; 
+			cav[cai++] = (char*)execpath; 
 			cav[cai++] = "-Fcgi";
 			cav[cai++] = cab;
 			cav[cai] = 0;
 			av = cav;
 		}
+		if( strcmp(shell,"#!/bin/sh") == 0
+		 && strcmp(BINSHELL,"/bin/sh") != 0 ){
+			cai = 0;
+			cav[cai++] = (char*)BINSHELL;
+			cav[cai++] = (char*)execpath;
+			for( ai = 1; av[ai]; ai++ ){
+				if( 32-1 <= cai ){
+					break;
+				}
+				cav[cai++] = av[ai];
+			}
+			cav[cai++] = 0;
+			av = cav;
+			execpath = BINSHELL;
+		}
 		fclose(cfp);
 	}
 
 	if( INHERENT_fork() ){
 		if( (pid = Fork("CGI")) == 0 ){
 		/* don't use Execvp() because it not returns even on error. */
-			execvp(execpath,av);
+			execvp(execpath,(char**)av);
 			putExecError(1,execpath);
 			Finish(-1);
 		}
@@ -486,48 +536,49 @@ cgi_process(Conn,tc,execpath,workdir,av,ev,pfp)
 
 	dup2(savfds[0],0); close(savfds[0]);
 	dup2(savfds[1],1); close(savfds[1]);
-	environ = savenv;
+	environ = (char**)savenv;
 	chdir(savdir);
 	pfp[1] = fdopen(toCGI[1],"w");
 	pfp[0] = fdopen(fromCGI[0],"r");
 	return pid;
 }
 
+int file_copyBuffered(FILE *in,FILE *out);
+int checkChildAbort1(PCStr(where));
+
 extern double HTTP_TOUT_QBODY;
-exec_cgi(Conn,req,reqhead,scriptpath,datapath,vurl,vpath,scripturl,extpath,fc,tc,stcodep)
-	Connection *Conn;
-	char *req,*reqhead,*scriptpath,*datapath;
-	char *vurl,*vpath,*scripturl,*extpath;
-	FILE *fc,*tc;
-	int *stcodep;
+int exec_cgi(Connection *Conn,PCStr(req),PCStr(reqhead),PCStr(scriptpath),PCStr(datapath),PCStr(vurl),PCStr(vpath),PCStr(scripturl),PCStr(extpath),FILE *fc,FILE *tc,int *stcodep)
 {
 	FILE *pfp[2];
-	char oreq[2048];
-	char tmp[128];
+	CStr(oreq,2048);
+	CStr(tmp,128);
 	int leng;
-	char workdir[1024];
-	char *tp;
-	char *av[32],*ev[128],eb[0x10000];
+	CStr(workdir,1024);
+	const char *tp;
+	const char *av[32]; /**/
+	const char *ev[128]; /**/
+	CStr(eb,0x10000);
 	StrVec Env;
-	char conninfo[4096];
+	CStr(conninfo,4096);
 	int pid;
+	FILE *xout;
 
-	make_conninfo(Conn,conninfo,NULL);
+	make_conninfo(Conn,AVStr(conninfo));
 	strcpy(workdir,datapath);
 	if( tp = strrchr(workdir,'/') )
-		*tp = 0;
-	av[0] = scriptpath;
+		truncVStr(tp);
+	av[0] = (char*)scriptpath;
 
-	SVinit(&Env,"exec_cgi",ev,128,eb,sizeof(eb));
+	SVinit(&Env,"exec_cgi",ev,elnumof(ev)-1,AVStr(eb),sizeof(eb)); /* -1 for the entry of randenv */
 	cgi_makeEnv(conninfo,req,reqhead,vurl,vpath,datapath,
-		scripturl,extpath, &av[1],&Env);
+		scripturl,extpath, 31,&av[1],&Env);
 	pid = cgi_process(Conn,tc,scriptpath,workdir,av,ev,pfp);
 
 	if( getFV(reqhead,"Content-Length",tmp) ){
 		int leng,wcc,wcc2,isbinary,timeout,ch;
 
 		leng = atoi(tmp);
-		timeout = HTTP_TOUT_QBODY;
+		timeout = (int)HTTP_TOUT_QBODY;
 		wcc = file_copyTimeout(fc,pfp[1],NULL,leng,&isbinary,timeout);
 		wcc2 = file_copyBuffered(fc,pfp[1]);
 		fflush(pfp[1]);
@@ -535,7 +586,11 @@ exec_cgi(Conn,req,reqhead,scriptpath,datapath,vurl,vpath,scripturl,extpath,fc,tc
 			wcc,wcc2,leng,isbinary?"(BINARY)":"");
 	}
 	fclose(pfp[1]);
+/*
 	leng = cgi_response(Conn,req,"",pfp[0],tc,stcodep);
+*/
+	xout = 0;
+	leng = cgi_response(Conn,req,CVStr(""),pfp[0],tc,&xout,stcodep);
 	fclose(pfp[0]);
 
 	if( 0 < pid ){
@@ -551,7 +606,10 @@ exec_cgi(Conn,req,reqhead,scriptpath,datapath,vurl,vpath,scripturl,extpath,fc,tc
 		usleep_bypoll(1000);
 	    }
 	    for( ntry = 0; ; ntry++ ){
+		xpid = checkChildAbort1("CGI");
+		/*
 		xpid = NoHangWait();
+		*/
 		sv1log("Wait*%d CGI-program exit: %d / %d (%d)\n",ntry,xpid,pid,
 			time(NULL)-start);
 		if( xpid == pid )
@@ -565,12 +623,36 @@ exec_cgi(Conn,req,reqhead,scriptpath,datapath,vurl,vpath,scripturl,extpath,fc,tc
 			usleep_bypoll(500*1000);
 	    }
 	}
+
+	if( xout != 0 ){
+		CStr(head,4096);
+		refQStr(hp,head); /**/
+		const char *xp;
+
+		sv1log("## SHTML from CGI\n");
+		fflush(xout);
+		fseek(xout,0,0);
+
+		xp = head + (sizeof(head)-1);
+		for(; hp < xp; hp += strlen(hp) ){
+			if( xp-hp < 1 )
+				break;
+			if( fgets(hp,xp-hp,xout) == NULL )
+				break;
+			if( *hp == '\r' || *hp == '\n' )
+				break;
+		}
+		leng = scan_SHTML(Conn,tc,fc,xout,req,head,vurl,vpath,
+			datapath,scripturl,extpath);
+		fclose(xout);
+	}
+
 	return leng;
 }
 
 /* sleep without SIGALRM ... to avoid the effect of -Tx ? */
 /*
-msleep_bypoll(msec)
+void msleep_bypoll(int msec)
 {	int waits[2];
 
 	Socketpair(waits);
@@ -580,10 +662,9 @@ msleep_bypoll(msec)
 }
 */
 
-static dump(request)
-	char *request;
+static void dump(PCStr(request))
 {	int ei;
-	char *env;
+	const char *env;
 
 	printf("Content-Type: text/html\r\n\r\n");
 	printf("<H2>DeleGate as a CGI program</H2>\n");
@@ -595,17 +676,29 @@ static dump(request)
 	fflush(stdout);
 }
 
-cgi_delegate(ac,av,Conn)
-	char *av[];
-	Connection *Conn;
+int cgi_delegate(int ac,const char *av[],Connection *Conn)
 {	char ei;
-	char request[4096],*rp;
-	char *env,*method,*url,*ver;
-	char field[128],value[4096];
-	char *host,*port,*path,server[1024],mount[1024],delegate[1024];
-	char *leng,*type;
-	char *query,qext[1024];
-	char *chost,*caddr,*cuser;
+	CStr(request,4096);
+	refQStr(rp,request); /**/
+	const char *env;
+	const char *method;
+	const char *url;
+	const char *ver;
+	CStr(field,128);
+	CStr(value,4096);
+	const char *host;
+	const char *port;
+	const char *path;
+	CStr(server,1024);
+	CStr(mount,1024);
+	CStr(delegate,1024);
+	const char *leng;
+	const char *type;
+	const char *query;
+	CStr(qext,1024);
+	const char *chost;
+	const char *caddr;
+	const char *cuser;
 	int fromHttpd[2],toHttpd[2];
 	int wcc;
 
@@ -622,11 +715,11 @@ cgi_delegate(ac,av,Conn)
 
 	if( method == 0 || host == 0 || port == 0 || path == 0 ){
 		printf("Status: 500 CGI-DeleGate Error\r\n");
-		return;
+		return -1;
 	}
 	if( url == 0 ){
 		printf("Status: 404 CGI-DeleGate Not Found\r\n");
-		return;
+		return -1;
 	}
 
 	/*
@@ -635,14 +728,15 @@ cgi_delegate(ac,av,Conn)
 	 */
 	if( strtailchr(path) != '/' )
 	if( url[0] == 0 ){
-				sprintf(url,"http://%s",host);
-				if( atoi(port) != 80 )
-					sprintf(url+strlen(url),":%s",port);
-				sprintf(url+strlen(url),"%s/",path);
-				printf("Status: 302 moved\r\n");
-				printf("Location: %s\r\n",url);
-				printf("\r\n");
-				return;
+		CStr(urlb,1024);
+		sprintf(urlb,"http://%s",host);
+		if( atoi(port) != 80 )
+		Xsprintf(TVStr(urlb),":%s",port);
+		Xsprintf(TVStr(urlb),"%s/",path);
+		printf("Status: 302 moved\r\n");
+		printf("Location: %s\r\n",urlb);
+		printf("\r\n");
+		return 0;
 	}
 	if( url[0] == 0 )
 		url = "/";
@@ -684,7 +778,6 @@ cgi_delegate(ac,av,Conn)
 	 * parent DeleGate??)
 	 */
 
-	rp = request;
 	if( query && *query )
 		sprintf(qext,"?%s",query);
 	else	qext[0] = 0;
@@ -701,14 +794,14 @@ if( strcmp(ver,"HTTP/1.1") == 0 ) ver = "HTTP/1.0";
 
 	for( ei = 0; env = environ[ei]; ei++ ){
 		if( strncmp(env,"HTTP_",5) == 0 )
-		if( sscanf(env+5,"%[^=]=%[^\n]",field,value) == 2 ){
-			char *fp;
+		if( Xsscanf(env+5,"%[^=]=%[^\n]",AVStr(field),AVStr(value)) == 2 ){
+			const char *fp;
 
 			for( fp = &field[1]; *fp; fp++ ){
 				if( isupper(*fp) )
-					*fp = tolower(*fp);
+					*(char*)fp = tolower(*fp);
 				if( *fp == '_' )
-					*fp = '-';
+					*(char*)fp = '-';
 			}
 			if( strcasecmp(field,"Accept") == 0 ){
 			}
@@ -756,7 +849,9 @@ if( strcmp(ver,"HTTP/1.1") == 0 ) ver = "HTTP/1.0";
 	 */
 	Socketpair(toHttpd);
 	if( Fork("CGI-DeleGate-From") == 0 ){
-		char stat[1024],*rcode,head1[4096];
+		CStr(stat,1024);
+		const char *rcode;
+		CStr(head1,4096);
 		FILE *resp;
 		int ch,hcc,bcc;
 
@@ -817,60 +912,136 @@ if( strcmp(ver,"HTTP/1.1") == 0 ) ver = "HTTP/1.0";
 	 */
 	wait(0);
 	wait(0);
+	return 0;
 }
 
-form2v(form,maxargc,argv)
-	char *form;
-	char *argv[];
-{	char *sp,*dp;
+int form2v(PVStr(form),int maxargc,const char *argv[])
+{	refQStr(sp,form); /**/
+	const char *dp;
 	int argc;
 
-	sp = form;
-	argc = 0;
-	for(;;){
-		argv[argc++] = sp;
-		if( dp = strchr(sp,'&') )
-			*dp++ = 0;
-		nonxalpha_unescape(sp,sp,1);
+	for( argc = 0; argc < maxargc; argc++ ){
+		if( maxargc-1 <= argc ){
+			daemonlog("F","## TOO MANY PARAMS IN FORM %d\n",argc);
+			break;
+		}
+		argv[argc] = sp;
+		if( dp = strchr(sp,'&') ){
+			truncVStr(dp); dp++;
+		}
+		nonxalpha_unescape(sp,AVStr(sp),1);
 		if( dp == 0 )
 			break;
-		sp = dp;
+		sp = (char*)dp;
 	}
 	argv[argc] = 0;
 	return argc;
 }
-HTTP_form2v(Conn,fc,maxargc,argv)
-	Connection *Conn;
-	FILE *fc;
-	char *argv[];
-{	char cLeng[1024],form[0x10000];
+int HTTP_form2v(Connection *Conn,FILE *fc,int maxargc,const char *argv[])
+{	CStr(cLeng,1024);
+	CStr(form,0x10000);
 	int cleng,rcc;
 	int argc,argi;
 
-	HTTP_getRequestField(Conn,"Content-Length",cLeng,sizeof(cLeng));
+	HTTP_getRequestField(Conn,"Content-Length",AVStr(cLeng),sizeof(cLeng));
 	cleng = atoi(cLeng);
 
 	argc = 0;
 	form[0] = 0;
 	if( 0 < cleng ){ /* && if Content-Type: x-form */
+		if( sizeof(form) <= cleng ){
+			daemonlog("F","## POST FORM TOO LARGE %d\n",cleng);
+			cleng = sizeof(form) -1;
+		}
 		rcc = fread(form,1,cleng,fc);
 		form[rcc] = 0;
-		argc = form2v(form,argc,argv);
+		argc = form2v(AVStr(form),maxargc,argv);
 	}
 	argv[argc] = 0;
 	return argc;
 }
 
-extern FILE *TMPFILE();
-system_CGI(conninfo,oreq,req,head,cgi,in,out)
-	char *conninfo,*oreq,*req,*head,*cgi;
-	FILE *in,*out;
-{	char *ev[128],eb[0x10000];
+int clientHTTP(Connection *Conn)
+{
+	if( strcaseeq(CLNT_PROTO,"http") || strcaseeq(CLNT_PROTO,"https") )
+		return CurEnv != 0;
+	return 0;
+}
+
+typedef struct {
+	StrVec	 e_Env;
+  const	char	*e_ev[128]; /**/
+	MStr(	 e_eb,0x10000);
+} CgiEnv;
+static const char **getCgiEnv(Connection *Conn,CgiEnv *E)
+{	CStr(ci,4096);
+	CStr(ourl,1024);
+
+	if( !clientHTTP(Conn) )
+		return (const char**)environ;
+
+	HTTP_originalURLx(Conn,AVStr(ourl),sizeof(ourl));
+	make_conninfo(Conn,AVStr(ci));
+	SVinit(&E->e_Env,"substCGIENV",E->e_ev,elnumof(E->e_ev)-1,AVStr(E->e_eb),sizeof(E->e_eb));
+	cgi_makeEnv(ci,OREQ,OREQ_MSG,"",ourl,REQ_URL,"","",0,NULL,&E->e_Env);
+	return E->e_ev;
+}
+int substCGIENV(Connection *Conn,PCStr(name),PVStr(out),int size)
+{	CgiEnv Ev;
+	const char *const*evp;
+	const char *e1;
+	int ei,nlen,match;
+
+	if( getenv("CFI_TYPE") ) /* running as a filter already with CGIENVs */
+		evp = (char const*const*)environ;
+	else	evp = getCgiEnv(Conn,&Ev);
+	nlen = strlen(name);
+	match = 0;
+	for( ei = 0; e1 = evp[ei]; ei++ ){
+		if( strncmp(name,e1,nlen) == 0 )
+		if( e1[nlen] == '=' ){
+			linescanX(e1+nlen+1,AVStr(out),size);
+			match = 1;
+		}
+	}
+	return match;
+}
+void putCGIENV(Connection *Conn)
+{	CgiEnv Ev;
+	const char *const*ev;
+
+	ev = getCgiEnv(Conn,&Ev);
+	if( ev != (char const*const*)environ )
+		environ = (char**)dupv(ev,0);
+}
+void pushCGIENV(Connection *Conn,void *sevp)
+{	char **ev;
+
+	ev = environ;
+	putCGIENV(Conn);
+	if( ev != environ )
+		*(char***)sevp = ev;
+	else	*(char***)sevp = 0;
+}
+void popCGIENV(Connection *Conn,void *sevp)
+{
+	if( *(char***)sevp && *(char***)sevp != environ ){
+		freev(environ);
+		environ = *(char***)sevp;
+	}
+}
+
+void System(PCStr(command),FILE *in,FILE *out);
+void system_CGI(PCStr(conninfo),PCStr(oreq),PCStr(req),PVStr(head),PCStr(cgi),FILE *in,FILE *out)
+{	const char *ev[128]; /**/
+	CStr(eb,0x10000);
 	FILE *tmp;
-	char *xhead,*dp;
-	char *url,*ourl;
+	const char *xhead;
+	const char *dp;
+	const char *url;
+	const char *ourl;
 	HttpRequest reqx,oreqx;
-	char **oenv;
+	const char *const *oenv;
 	StrVec Env;
 
 	xhead = stralloc(head);
@@ -878,22 +1049,22 @@ system_CGI(conninfo,oreq,req,head,cgi,in,out)
 	ourl = oreqx.hq_url;
 	decomp_http_request(req,&reqx);
 	url = reqx.hq_url;
-	SVinit(&Env,"sysgem_CGI",ev,128,eb,sizeof(eb));
-	cgi_makeEnv(conninfo,req,xhead,"",ourl,url,"","",NULL,&Env);
+	SVinit(&Env,"sysgem_CGI",ev,elnumof(ev)-1,AVStr(eb),sizeof(eb));
+	cgi_makeEnv(conninfo,req,xhead,"",ourl,url,"","",0,NULL,&Env);
 
-	oenv = environ;
-	environ = ev;
+	oenv = (char const*const*)environ;
+	environ = (char**)ev;
 
 	tmp = TMPFILE("CFI-system_CGI");
 	System(cgi,in,tmp);
-	environ = oenv;
+	environ = (char**)oenv;
 	fseek(tmp,0,0);
 	if( dp = strstr(head,"\r\n\r\n") )
-		dp[2] = 0;
+		((char*)dp)[2] = 0;
 	if( dp = strstr(head,"\n\n") )
-		dp[1] = 0;
-	cgi_response(NULL,req,head,tmp,out,NULL);
+		((char*)dp)[1] = 0;
+	cgi_response(NULL,req,AVStr(head),tmp,out,NULL,NULL);
 	fclose(tmp);
 
-	free(xhead);
+	free((char*)xhead);
 }
